# Author: Matouš Elphick
# License: BSD 3 clause
import numpy as np
import inspect
import pyclesperanto_prototype as cle
from apoc._feature_sets import PredefinedFeatureSet
from scipy import stats, ndimage
from skimage.segmentation import watershed
from sklearn.metrics import jaccard_score
from sklearn.preprocessing import LabelEncoder
from typing import Union, Optional


def mask_img(image, mask):
    """
    Applies binary mask to image.

    Args:
        image: Image to apply binary mask to.
        mask: Binary mask.

    Returns:
        Image with binary mask applied.
    """
    mask[mask > 0] = 1
    return image * mask


def remove_membrane(image_input, mask_input, compactness: int = 0.005, verbose=False):
    """
    This function removes the membrane and generates
    internal mask and applies it to the raw input image
    for the "Complete Process" execution mode.

    Uses sobel filter -> watershed -> local mean neighbor count -> fancy indexing

    Args:
        verbose:
        compactness: Higher values result in more regularly-shaped watershed basins.
        image_input: Raw Image
        mask_input: Mitochondria Mask Image

    Returns:
        Generate the internal mask and applies it to the raw image.
    """
    processed = []
    for image_index in range(len(image_input)):
        sobel = cle.pull(cle.sobel(image_input[image_index])).astype(np.uint32)
        segments_watershed = watershed(sobel, compactness=compactness, mask=mask_input[image_index])
        touch_matrix = cle.generate_touch_matrix(segments_watershed)
        neighbor_count = cle.count_touching_neighbors(touch_matrix)
        local_mean_neighbor_count = cle.mean_of_touching_neighbors(neighbor_count, touch_matrix)
        parametric_image = cle.replace_intensities(segments_watershed, local_mean_neighbor_count)
        membrane_mask = cle.pull(parametric_image).astype(np.uint32)
        membrane_mask[membrane_mask <= 4] = 0
        membrane_mask[membrane_mask == stats.mode(membrane_mask[0], keepdims=False)[0]] = 0
        membrane_mask[membrane_mask > 0] = 1
        processed.append(image_input[image_index] * np.array(ndimage.binary_fill_holes(membrane_mask)))
    return np.asarray(processed)


def to_np(features, label=None):
    """
    Function which takes a feature set and or ground truth
    and reshapes it such that it can be used as training
    and labels for the VEMClassifier.

    Args:
        features: Image X features
        label: Image y labels

    Returns:
        `X`, `y` training data in format readable by VEMClassifier
        and the `feature_stack`.
    See Also:
        https://github.com/haesleinhuepf/apoc
    """
    feature_stack = features.reshape(-1, features.shape[3])
    if label is None:
        return feature_stack
    else:
        # make the annotation 1-dimensional
        ground_truth_np = label.reshape(-1)

        X = feature_stack
        y = ground_truth_np

        # remove all pixels from the feature and annotations which have not been annotated
        mask = y > 0
        X = X[mask]
        y = y[mask]

        # Encode target labels with value between 0 and n_classes-1.
        le = LabelEncoder()
        y = le.fit_transform(y)

        return X, y, feature_stack


def pre_process(X, mask=None, rem_membrane=False, verbose=False):
    """
    Function to check if there is to be any preprocessing
    on the user inputted X. If mask is not none then a mask
    will be applied. If the mask is not none and rem_membrane
    is True then the outer segments of the mask will be removed.

    Args:
        verbose:
        X: Image
        mask: Mask image
        rem_membrane: Flag to remove membrane

    Returns:
        X which has had preprocessing applied.
    """
    if mask is not None and rem_membrane:
        if mask.any():
            X = remove_membrane(X, mask, verbose)
    elif mask is not None and not rem_membrane:
        if mask.any():
            X = mask_img(X, mask)

    return X


def img_to_3d(img) -> np.array:
    """
    Simple function to reformat 2D image to
    a single 3D Image. For example an image of
    input (255 x 255) will be returned as
    (1 x 255 x 255).


    Args:
        img: Image to convert

    Returns:
        Image with added dimension
    """
    img = np.array(img)
    return img.reshape((1,) + img.shape)


def generate_feature_stack(image, features_specification: Union[str, PredefinedFeatureSet] = None):
    """
    Creates a feature stack from a given image.

    Parameters
    ----------
    image : ndarray
        2D or 3D image to generate a feature stack from
    features_specification : str or PredefinedFeatureSet
        a space-separated list of features, e.g.
        original gaussian=4 sobel_of_gaussian=4 or a
        PredefinedFeatureSet

    See Also
    --------
    https://github.com/haesleinhuepf/apoc

    Returns
    -------
    a list of OCLarray images
    """

    image = cle.push(image)

    # default features
    if features_specification is None:
        blurred = cle.gaussian_blur(image, sigma_x=2, sigma_y=2, sigma_z=2)
        edges = cle.sobel(blurred)
        stack = [
            image,
            blurred,
            edges
        ]

        return stack
    if isinstance(features_specification, PredefinedFeatureSet):
        features_specification = features_specification.value

    while "  " in features_specification:
        features_specification = features_specification.replace("  ", " ")
    while "\t" in features_specification:
        features_specification = features_specification.replace("\t", " ")

    features_specs = features_specification.split(" ")
    generated_features = {}

    result_features = []

    for spec in features_specs:
        if spec.lower() == 'original':
            generated_features['original'] = image
            result_features.append(image)
        elif "=" in spec:
            temp = spec.split("=")
            operation = temp[0]
            numeric_parameter = float(temp[1])

            if not hasattr(cle, operation) and "_of_" in operation:
                temp = operation.split("_of_")
                outer_operation = temp[0]
                inner_operation = temp[1]

                if (inner_operation + "=" + str(numeric_parameter)) not in generated_features.keys():
                    new_image = cle.create_like(image)
                    _apply_operation(inner_operation, image, new_image, numeric_parameter)
                    generated_features[inner_operation + "=" + str(numeric_parameter)] = new_image

                if (operation + "=" + str(numeric_parameter)) not in generated_features.keys():
                    new_image2 = cle.create_like(image)
                    _apply_operation(outer_operation,
                                     generated_features[inner_operation + "=" + str(numeric_parameter)], new_image2,
                                     numeric_parameter)
                    generated_features[operation + "=" + str(numeric_parameter)] = new_image2
            else:
                if (operation + "=" + str(numeric_parameter)) not in generated_features:
                    new_image = cle.create_like(image)
                    _apply_operation(operation, image, new_image, numeric_parameter)
                    generated_features[operation + "=" + str(numeric_parameter)] = new_image
            result_features.append(cle.pull(generated_features[operation + "=" + str(numeric_parameter)]))
        elif not hasattr(cle, spec.lower()):
            raise (AttributeError("module 'pyclesperanto_prototype' has no attribute '{}'".format(spec)))

    result_features = np.moveaxis(result_features, 0, -1)
    if not np.any(result_features):
        raise MemoryError("Dataset exceeded memory, please use a subset of the dataset")

    return result_features


def _apply_operation(operation, input_image, output_image, numeric_parameter):
    """Apply a given image-filter to an image and save the result into another new_image.

    Parameters
    ----------
    operation: callable
    input_image: ndimage
    output_image: ndimage
    numeric_parameter: float or int
        The filters typically have numeric parameters, such as radius or sigma.

    See Also
    --------
    https://github.com/haesleinhuepf/apoc
    """
    func = getattr(cle, operation)
    sig = inspect.signature(func)
    if len(sig.parameters.keys()) == 2:
        func(input_image, output_image)
    elif len(sig.parameters.keys()) == 3:
        func(input_image, output_image, numeric_parameter)
    elif len(sig.parameters.keys()) == 4:
        func(input_image, output_image, numeric_parameter, numeric_parameter)
    elif len(sig.parameters.keys()) == 5:
        func(input_image, output_image, numeric_parameter, numeric_parameter, numeric_parameter)
    elif len(sig.parameters.keys()) == 8:
        # e.g. difference_of_gaussian
        func(input_image, output_image, numeric_parameter * 0.9, numeric_parameter * 0.9, numeric_parameter * 0.9,
             numeric_parameter * 1.1, numeric_parameter * 1.1, numeric_parameter * 1.1)
    else:
        func(input_image, output_image, numeric_parameter, numeric_parameter, numeric_parameter)


def predict_iou(y_true,
                y_pred,
                average: Optional[str] = 'binary'
                ):
    """

    Args:
        y_true:
        y_pred:
        average:

    Returns:

    """

    return jaccard_score(y_true.flatten(), y_pred.flatten(), average=average)



# Author: Matouš Elphick
# License: BSD 3 clause
import matplotlib.pyplot as plt
import numpy as np
import pyclesperanto_prototype as cle
from typing import Any, Optional, Sequence, Tuple, Union
from sklearn.model_selection import GridSearchCV
from xgboost import XGBClassifier, XGBModel
from xgboost.callback import TrainingCallback
from xgboost.core import Booster, ArrayLike
from ._utils import to_np, pre_process, generate_feature_stack


class VEMClassifier(XGBClassifier):
    def __init__(self,
                 *,
                 features: Optional[str] = 'original',
                 verbosity: Optional[int] = 0,
                 **kwargs: Any):
        """
        A XGBoost classifier with added functionality optimised
        for Volume Electron Microscopy data.

        Args:
            **kwargs: Parameters for XGBoost classifier.

        See Also:
            https://xgboost.readthedocs.io/en/stable/python/python_api.html#xgboost.XGBClassifier
        """
        self.features = features
        super().__init__(verbosity=verbosity, **kwargs)

    def plot_feature_importance(self, fig_size: Optional[tuple] = (10, 5)):
        """
        Function to plot the feature importance after training is completed.

        Args:
            fig_size: Size of the figure to be outputted.

        Returns:
            Figure showing the importance of features in descending order.
        """

        feature_names = self.features.split(' ')
        sorted_idx = super().feature_importances_.argsort()

        plt.figure(figsize=fig_size)
        plt.barh([feature_names[i] for i in sorted_idx], super().feature_importances_[sorted_idx])
        plt.xlabel("Xgboost Feature Importance")
        plt.tight_layout()
        plt.show()

    def GridSearchCV(self,
                     X: ArrayLike,
                     y: ArrayLike,
                     param_grid: dict,
                     *,
                     mask: Optional[ArrayLike] = None,
                     rem_membrane: Optional[bool] = False,
                     update_params: Optional[bool] = False,
                     **kwargs
                     ) -> "GridSearchCV":
        """
        Function to use sklearn GridSearchCV on
        an initialised VEMClassifier and optionally
        update the parameters of the VEMClassifier using
        the best parameters from the GridSearchCV.

        Args:
            X: Image feature matrix.
            y: Image Labels.
            param_grid: dict or list of dictionaries. Dictionary with
                parameters names (`str`) as keys and lists of parameter
                settings.
            mask: Apply binary mask to X.
            rem_membrane:  Remove membrane from X after applying mask.
            update_params: Option to take the parameters from GridSearch best
                VEMClassifier and update the current parameters.
            **kwargs: Custom arguments for the GridSearchCV

        Returns:
            Resulting GridSearchCV object.
        See Also:
            https://scikit-learn.org/stable/modules/generated/sklearn.model_selection.GridSearchCV.html
        """

        # If feature set is already defined
        if X.ndim != 2 and y.ndim != 1:
            X = pre_process(X, mask=mask, rem_membrane=rem_membrane)

            train_features = np.asarray(generate_feature_stack(X, self.features))

            X_test, y_test, _ = to_np(train_features, y)
        else:
            X_test = X
            y_test = y

        grid_search = GridSearchCV(
            estimator=self,
            param_grid=param_grid,
            **kwargs,
        )

        grid_search.fit(X_test, y_test)

        if update_params:
            params = grid_search.best_estimator_.get_params()
            super().set_params(**params)

        return grid_search

    def fit(
            self,
            X: ArrayLike,
            y: ArrayLike,
            *,
            verbose: Optional[Union[bool, int]] = False,
            xgb_model: Optional[Union[Booster, str, XGBModel]] = None,
            callbacks: Optional[Sequence[TrainingCallback]] = None,
            mask: Optional[ArrayLike] = None,
            rem_membrane: bool = False,
    ) -> "XGBClassifier":
        """
        Function to fit VEMClassifier and optionally remove mask the
        inputs X dnd Y, and or remove the membrane.

        Note that calling fit() multiple times will cause the model
        object to be re-fit from scratch. To resume training from a
        previous checkpoint, explicitly pass xgb_model argument.

        Args:
            X: Image feature matrix.
            y: Image labels.
            verbose: If verbose is True and an evaluation set is used, the evaluation
                metric measured on the validation set is printed to stdout at each
                boosting stage. If verbose is an integer, the evaluation metric is
                printed at each verbose boosting stage. The last boosting stage /
                the boosting stage found by using early_stopping_rounds is also printed.
            xgb_model: file name of stored XGBoost model or ‘Booster’ instance XGBoost
                model to be loaded before training (allows training continuation).
            callbacks: Deprecated since version 1.6.0: Use `callbacks` in `__init__()` or
                `set_params()` instead.
            mask: Apply binary mask to X.
            rem_membrane: Remove membrane from X after applying mask.

        Returns:
            VEMClassifier
        See Also:
            https://xgboost.readthedocs.io/en/stable/python/python_api.html#xgboost.XGBClassifier.fit
        """

        if X.ndim == 2 and y.ndim == 1:
            return super().fit(X,
                               y,
                               sample_weight=None,
                               base_margin=None,
                               eval_set=None,
                               early_stopping_rounds=None,
                               verbose=verbose,
                               xgb_model=xgb_model,
                               sample_weight_eval_set=None,
                               base_margin_eval_set=None,
                               eval_metric=None,
                               feature_weights=None,
                               callbacks=callbacks)

        X = pre_process(X, mask=mask, rem_membrane=rem_membrane)
        train_features = generate_feature_stack(X, self.features)

        X_train, y_train, _ = to_np(train_features, y)

        return super().fit(X_train,
                           y_train,
                           sample_weight=None,
                           base_margin=None,
                           eval_set=None,
                           early_stopping_rounds=None,
                           verbose=verbose,
                           xgb_model=xgb_model,
                           sample_weight_eval_set=None,
                           base_margin_eval_set=None,
                           eval_metric=None,
                           feature_weights=None,
                           callbacks=callbacks)

    def predict(
            self,
            X: ArrayLike,
            *,
            output_margin: bool = False,
            validate_features: bool = True,
            iteration_range: Optional[Tuple[int, int]] = None,
            show: Optional[bool] = False,
            mask: Optional[np.ndarray] = None,
            rem_membrane: Optional[bool] = False
    ) -> np.ndarray:
        """
        Predict with X and optionally remove mask the inputs X dnd Y,
        and or remove the membrane. If the model is trained with early stopping,
        then best_iteration is used automatically. For tree models, when data
        is on GPU, like cupy array or cuDF dataframe and predictor is not
        specified, the prediction is run on GPU automatically,
        otherwise it will run on CPU.

        Args:
            X: Image feature matrix.
            output_margin: Whether to output the raw untransformed margin value.
            validate_features: When this is True, validate that the Booster’s
                and data’s feature_names are identical. Otherwise, it is assumed
                that the feature_names are the same.
            iteration_range: Specifies which layer of trees are used in prediction.
                For example, if a random forest is trained with 100 rounds. Specifying
                 `iteration_range=(10, 20)`, then only the forests built during [10, 20)
                 (half open set) rounds are used in this prediction.
            show: Output an image of the predictions.
            mask: Apply binary mask to X.
            rem_membrane: Remove membrane from X after applying mask.

        Returns:
            VEMClassifier prediction.
        See Also:
            https://xgboost.readthedocs.io/en/stable/python/python_api.html#xgboost.XGBClassifier.predict
        """

        if X.ndim == 2:
            predictions = super().predict(X,
                                          ntree_limit=None,
                                          output_margin=output_margin,
                                          validate_features=validate_features,
                                          base_margin=None,
                                          iteration_range=iteration_range)
            if show:
                cle.imshow(predictions.reshape(X.shape), labels=True)

            return predictions

        processed = pre_process(X, mask=mask, rem_membrane=rem_membrane)
        X_features = generate_feature_stack(processed, self.features)
        X_predict = to_np(X_features)

        predictions = super().predict(X_predict,
                                      ntree_limit=None,
                                      output_margin=output_margin,
                                      validate_features=validate_features,
                                      base_margin=None,
                                      iteration_range=iteration_range)
        predictions += 1

        if show:
            cle.imshow(predictions.reshape(X.shape), labels=True)

        return predictions.reshape(X.shape)

    def predict_proba(
            self,
            X: ArrayLike,
            validate_features: bool = True,
            iteration_range: Optional[Tuple[int, int]] = None,
            mask=None,
            rem_membrane: bool = False,
    ) -> np.ndarray:
        """
        Predict the probability of each `X` example being of a given class.
        Optionally remove mask the inputs X dnd Y, and or remove the membrane.

        Args:
            X: Image feature matrix.
            validate_features: When this is True, validate that the Booster’s
                and data’s feature_names are identical. Otherwise, it is assumed
                that the feature_names are the same.
            iteration_range: Specifies which layer of trees are used in prediction.
                For example, if a random forest is trained with 100 rounds. Specifying
                 `iteration_range=(10, 20)`, then only the forests built during [10, 20)
                 (half open set) rounds are used in this prediction.
            mask: Apply binary mask to X.
            rem_membrane: Remove membrane from X after applying mask.

        Returns:
            prediction :
            a four dimensional numpy array of shape array-like of shape (n_classes, Z, X, Y)
            with the probability of each data example being of a given class.
        """
        if X.ndim == 2:
            predictions = super().predict_proba(X,
                                                ntree_limit=None,
                                                validate_features=validate_features,
                                                base_margin=None,
                                                iteration_range=iteration_range)
            return predictions

        X = pre_process(X, mask=mask, rem_membrane=rem_membrane)

        X_features = generate_feature_stack(X, self.features)

        X_predict = to_np(X_features)

        predictions = super().predict_proba(X_predict,
                                            ntree_limit=None,
                                            validate_features=validate_features,
                                            base_margin=None,
                                            iteration_range=iteration_range)
        n_classes = predictions.shape[1]
        flattened_prob = []
        for i in range(n_classes):
            flattened_prob.append(predictions[:, i].reshape(X.shape))
        flattened_prob = np.array(flattened_prob)
        return flattened_prob



import setuptools

with open('README.md') as f:
    long_description = f.read()

setuptools.setup(
    name='vemseg',
    version='0.0.5',
    packages=['vemseg'],
    url='https://github.com/MatousE/vemseg',
    license='BSD 3-Clause License',
    author='MatousE',
    author_email='matous.elphick@gmail.com',
    description='A python package for the segmentation of organelles in volume electron microscopy data',
    long_description=long_description,
    long_description_content_type='text/markdown',
    include_package_data=True,
    install_requires=['numpy',
                      'apoc',
                      'scipy',
                      'scikit-image',
                      'scikit-learn',
                      'matplotlib',
                      'xgboost',
                      'setuptools'
                      ]
)
